"""ttcpcontrol

A some utility classes useful for controlling the execution of instances of the
ADI-TTCP network throughput test program and for obtaining the results.

The CtrlStream class makes explicit use of Tkinter widgets in order to display
a top-level PanedWindow in which it logs activity on the control stream.

"""

import sys
import time
import asynchat, asyncore, socket
import Tkinter      as Tk
import tkFont       as Tf
import tkMessageBox as Tm
import ScrolledText as St
import re

# protocol identifiers
TCP = 0
UDP = 1

# the default TCP port on which ADI-TTCP listens for control connections
DEFAULT_CONTROL_PORT = 6000

# default values for Transfer attributes
DEFAULT_DATA_PORT    = 7000     # receive port (TCP and UDP); can only be used
                                # once per protocol per batch of Transfers
DEFAULT_PROTOCOL     = TCP      # the protocol to use
DEFAULT_BUFLEN       = 1000     # the size of buffer to write or read
DEFAULT_NUMBUFS      = 20000    # the number of buffers to send
DEFAULT_DELAY        = 0        # microseconds of delay between writes
DEFAULT_RATE         = 0        # desired transmit rate in bytes/sec
DEFAULT_PERIOD       = 0        # number of seconds for which to write
DEFAULT_CONTINUOUS   = False    #
DEFAULT_TCP_NODELAY  = False    # set TCP_NODELAY socket option?
DEFAULT_SO_DEBUG     = False    # set SO_DEBUG socket option?
DEFAULT_SO_SNDBUF    = 0        # set SO_SNDBUF socket option?
DEFAULT_SO_RCVBUF    = 0        # set SO_RCVBUF socket option?
DEFAULT_ALIGN        = 0        # buffer alignment
DEFAULT_OFFSET       = 0        # buffer offset
DEFAULT_FORMAT       = 'K'      # throughput reporting unit
DEFAULT_TOUCH        = False    # touch received data?

_errorfunc = None # function to call when it all goes wrong
_errorbox  = None # error message box instance

#-------------------------------------------------------------------------------

class CtrlStream(asynchat.async_chat):

    """Implements communication over an ADI-TTCP instance's TCP control stream.

    A script should create one instance of CtrlStream for each instance of the
    ADI-TTCP program with which it wishes to communicate. It should use the
    write() and read() methods defined here to perform the communication rather
    than async_chat's methods so that traffic will be written to the log window.

    """

    logw = None # TopLevel for log window
    pane = None # PanedWindow into which a ScolledText is fitted for every
                # live CtrlStream
    font = None # standard tkFont to use for the ScrolledText windows
    bold = None # bold tkFont to use for heading

    plid = None # call-back ID for poll()

    def poll():
        """Must be called regularly to poll for input.
        """
        CtrlStream.plid = None
        asyncore.loop(0.0, True, None, 1)
        CtrlStream.plid = CtrlStream.logw.after(500, CtrlStream.poll)

    # make poll() a static method
    poll = staticmethod(poll)

    def init(error_func):
        global _errorfunc
        _errorfunc = error_func

        logw = Tk.Toplevel()
        logw.title("ADI-TTCP Logs")
        logw.resizable(1,1)
        logw.geometry('560x683+10+10')
        logw.protocol('WM_DELETE_WINDOW', CtrlStream.fini)
        CtrlStream.logw = logw

        pane = Tk.PanedWindow(logw, orient='vertical')
        pane.pack(expand=1, fill='both')
        CtrlStream.pane = pane

        CtrlStream.font = Tf.Font(logw, ('Courier New', 8, Tf.NORMAL))
        CtrlStream.bold = Tf.Font(logw, ('Courier New', 9, Tf.BOLD))

        # initiate polling for control stream activity
        CtrlStream.plid = logw.after(1000, CtrlStream.poll)

    # make init() a static method
    init = staticmethod(init)

    def fini():
        if CtrlStream.plid: CtrlStream.logw.after_cancel(CtrlStream.plid)
        CtrlStream.logw.destroy()

    # make fini() a static method
    fini = staticmethod(fini)

    def __init__(self, host, port=DEFAULT_CONTROL_PORT, descr="", relay=None):
        """Constructor.

        Calls the async_chat constructor with default arguments and then creates
        a TCP socket and attempts to connect to the address specified by the
        'host' and 'port' arguments.

        Optional argument 'descr' is a string that is intended to describe the
        platform on which the controlled ADI-TTCP program is running, perhaps
        naming the CPU type and giving relevant clock speeds, etc.

        If optional argument 'relay' is supplied and is not None then it must
        be an instance of CtrlStream. It will be sent a RELAY command for 'host'
        and 'port' and the CtrlStream being constructed here will open its
        socket on 'relay's IP address and the port number returned by the
        RELAY command.

        The constructor also opens a new pane in the log window for recording
        traffic on this control stream.
        """
        asynchat.async_chat.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)

        if relay == None:
            # a straightforward connection to (host, port)
            self.connect((host, port))
        else:
            # target is to be reached via 'relay'
            relay.write('RELAY %s %d\r\n' % (host, port))
            rport = int(relay.read())
            self.connect((relay.host, rport))

        self.relay = relay

        self.set_terminator('\n')
        self.data    = ""
        self.host    = host
        self.port    = port
        self.descr   = descr
        self.address = "%s:%d" % (host, port)
        self.lframe  = Tk.LabelFrame(CtrlStream.pane, text=self.address)
        self.txt = St.ScrolledText(self.lframe,
                                   state='disabled',
                                   relief='sunken',
                                   pady=5,
                                   font=CtrlStream.font
                                  )
        self.txt.pack(expand=1, fill='both')
        CtrlStream.pane.add(self.lframe, minsize='1i', height='200')
        self.txt.config(state='normal')
        msg = "ADI-TTCP Control Stream Log for %s\n" % self.address
        self.txt.insert(Tk.END, msg, 'htag')
        self.txt.tag_config('htag', font=CtrlStream.bold, justify='center')
        msg = "[%s]\n" % self.descr
        self.txt.insert(Tk.END, msg, 'dtag')
        self.txt.tag_config('dtag', justify='center')
        self.txt.insert(Tk.END, '\n')
        self.txt.config(state='disabled')

        self.fifo = asynchat.fifo()
        self.avail = Tk.IntVar(self.txt)
        self.avail.set(0)

        self.numResults = 0

    def getControlInfo(self):
        """ Return information about this CtrlStream instance.

        Returns a tuple of the host, port and description associated with
        this instance of CtrlStream.
        """
        return (self.host, self.port, self.descr)

    def collect_incoming_data(self, data):
        data = data.strip('\r')
        self.txt.config(state='normal')
        self.txt.insert(Tk.END, data)
        self.txt.config(state='disabled')
        self.data += data

    def found_terminator(self):
        self.txt.config(state='normal')
        self.txt.insert(Tk.END, '\n')
        self.txt.see(Tk.END)
        self.txt.config(state='disabled')
        self.data += '\r\n'
        if len(self.fifo) == 0:
            self.fifo.push(self.data)
            self.avail.set(~self.avail.get())
        else:
            self.fifo.push(self.data)
        self.data = ""

    def write(self, data):
        self.txt.config(state='normal')
        self.txt.insert(Tk.END, data[0:data.rindex('\r')] + '\n', 'xtag')
        self.txt.tag_config('xtag', background='LightBlue2')
        self.txt.see(Tk.END)
        self.txt.config(state='disabled')
        self.push(data)

    def read(self):
        if len(self.fifo) == 0:
            self.txt.wait_variable(self.avail)
        return self.fifo.pop()[1]

    def read_until(self, marker):
        txt  = ""
        line = ""
        while line.find(marker) == -1:
            line = self.read()
            txt += line
        return txt

    def handle_connect(self):
        pass

    def handle_close(self):
        self.close()
        CtrlStream.pane.remove(self.lframe)

    def handle_expt(self):
        global _errorfunc, _errorbox
        if _errorbox == None:
            msg = 'Failed to communicate with control stream at %s.\n' % \
                  self.address
            msg += 'Please ensure ADI-TTCP is running on that machine.\n'
            print msg
            _errorbox = Tm.showerror('ADI-TTCP Control Stream Error',msg)
            _errorfunc()

#-------------------------------------------------------------------------------

class Transfer(object):

    """ADI-TTCP Transfer class.

    A Transfer instance holds sufficient information about a throughput test
    to allow it to issue appropriate TRANSFER commands to the two instances
    of the ADI-TTCP program involved in the test.

    A script should create one instance of Transfer for each throughput test
    that is to be run concurrently and then call static function Run() to
    execute the tests and obtain the results.

    The Transfer constructor requires two positional arguments, instances of
    CtrlStream for the transmit and receive end respectively, and accepts
    any of the following optional keyword arguments:

        port       = integer: port number for receive end of transfer
        proto      = integer: ttcplib.TCP or ttcplib.UDP
        buflen     = integer: size of writes to and reads from socket
        numbufs    = integer: number of buffers to write
        delay      = integer: microseconds of delay to insert between writes
        rate       = integer: desired transfer rate (in bytes per second)
        period     = integer: number of seconds for which to transmit
        continuous =
        nodelay    = boolean: whether or not to set TCP_NODELAY socket option
        debug      = boolean: whether or not to set SO_DEBUG socket option
        sndbuf     = integer: size for SO_SNDBUF socket option
        rcvbuf     = integer: size for SO_RCVBUF socket option
        align      = integer: alignment for send and receive buffers
        offset     = integer: offset within buffer at which to send/receive
        format     = string:  unit for throughput message ('M', 'K', etc)
        touch      = boolean: whether or not to 'touch' each received byte

    """

    batch = list()  # list of Transfers to be processed by Run()

    def __init__(self,
                 ctrl1,     # transmitter's Controller instance
                 ctrl2,     # receiver's Controller instance
                 **kwargs):

        # set up default transfer attributes
        self.ctrl1      = ctrl1
        self.ctrl2      = ctrl2
        self.port       = DEFAULT_DATA_PORT
        self.proto      = DEFAULT_PROTOCOL
        self.buflen     = DEFAULT_BUFLEN
        self.numbufs    = DEFAULT_NUMBUFS
        self.delay      = DEFAULT_DELAY
        self.rate       = DEFAULT_RATE
        self.period     = DEFAULT_PERIOD
        self.continuous = DEFAULT_CONTINUOUS
        self.nodelay    = DEFAULT_TCP_NODELAY
        self.debug      = DEFAULT_SO_DEBUG
        self.sndbuf     = DEFAULT_SO_SNDBUF
        self.rcvbuf     = DEFAULT_SO_RCVBUF
        self.align      = DEFAULT_ALIGN
        self.offset     = DEFAULT_OFFSET
        self.format     = DEFAULT_FORMAT
        self.touch      = DEFAULT_TOUCH

        # now update attributes from the keyword arguments (if any)
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                print "Warning: unrecognised transfer attribute '" + \
                      key + \
                      "' ignored"

        # initialise runtime attributes
        self.xmtID          = 0         # TRANSFER -t identifier
        self.xmtTxt         = ""        # ... and result string
        self.rcvID          = 0         # TRANSFER -r identifier
        self.rcvTxt         = ""        # ... and result string

        # add this new Transfer to the current batch
        Transfer.batch.append(self)


    def Run():
        """Run() static method

        Iterates through the current batch of Transfers constructing a transmit
        and receive TRANSFER command for each and sending them to appropriate
        ADI-TTCP control streams. Then issues RUN command to all control streams
        and waits for all the transfers to complete.

        Returns the list of Transfers in the batch. The xmtTxt and rcvTxt
        members of each Transfer are strings containing the output returned
        from the transmit and receive sides of the transfer.  The output strings
        have been stripped of their "START RESULTS" / "END RESULTS" markers.

        """
        # set of control instances used in this batch
        controllers = set()

        # dictionary mapping CtrlStream and transfer ID to Transfer
        tmap = dict()

        # construct and send TRANSFER commands for each Transfer in batch
        for t in Transfer.batch:
            xmtHost, xmtPort, xmtDescr = t.ctrl1.getControlInfo()
            rcvHost, rcvPort, rcvDescr = t.ctrl2.getControlInfo()

            # transmit side
            cmd  = "TRANSFER -t -p %d " % t.port
            if t.proto == UDP:cmd += "-u "
            if t.buflen > 0:  cmd += "-l %d " % (t.buflen)
            if t.delay > 0:   cmd += "-w %d " % (t.delay)
            if t.rate > 0:    cmd += "-L %d " % (t.rate)
            if t.period > 0:  cmd += "-P %d " % (t.period)
            if t.continuous or t.period > 0:
                cmd += "-c "
            else:
                cmd += "-n %d " % (t.numbufs)
            if t.nodelay:     cmd += "-D "
            if t.debug:       cmd += "-d "
            if t.sndbuf > 0:  cmd += "-b %d " % (t.sndbuf)
            if t.align > 0:   cmd += "-A %d " % (t.align)
            if t.offset > 0:  cmd += "-O %d " % (t.offset)
            cmd += "-f %s %s\r\n" % (t.format, rcvHost)

            t.ctrl1.write(cmd)

            # receive side
            cmd  = "TRANSFER -r -p %d " % t.port
            if t.proto == UDP:cmd += "-u "
            if t.buflen > 0:  cmd += "-l %d " % (t.buflen)
            if t.nodelay:     cmd += "-D "
            if t.debug:       cmd += "-d "
            if t.rcvbuf > 0:  cmd += "-b %d " % (t.rcvbuf)
            if t.align > 0:   cmd += "-A %d " % (t.align)
            if t.offset > 0:  cmd += "-O %d " % (t.offset)
            if t.touch:       cmd += "-T "
            cmd += "-f %s\r\n" % (t.format)

            t.ctrl2.write(cmd)

            # obtain the TRANSFER identifiers for each side
            txt = t.ctrl1.read()
            t.xmtID = int(txt)
            txt = t.ctrl2.read()
            t.rcvID = int(txt)

            # add transmit and receive side Controllers to set of
            # unique Controllers used in this batch
            controllers.add(t.ctrl1)
            controllers.add(t.ctrl2)

            # add entries to Transfer map
            tmap[(t.ctrl1,t.xmtID)] = (t, 'T')
            tmap[(t.ctrl2,t.rcvID)] = (t, 'R')

            # increment expected results counts
            t.ctrl1.numResults += 1
            t.ctrl2.numResults += 1

        # send a RUN command to every CtrlStream involved in this batch
        for c in controllers:
            c.write("RUN\r\n")

        # obtain output from each end of each transfer
        for c in controllers:
            while c.numResults > 0:
                txt = c.read_until('END RESULTS')
                id = int(txt[0:txt.index(' ')])
                s = txt.index('\r\n') + 2
                e = txt.rindex('\r\n', 0, len(txt)-3)
                txt = txt[s:e]
                t, dir = tmap[(c,id)]
                if dir == 'T':
                    t.xmtTxt = txt
                else:
                    t.rcvTxt = txt
                c.numResults -= 1


        tfrlist = Transfer.batch

        # empty the list of current transfers
        Transfer.batch = list()

        # return results
        return tfrlist

    # make Run() a static method
    Run = staticmethod(Run)

#-------------------------------------------------------------------------------
